/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"
#include <X11/Xos.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/Frame.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/Label.h>
#include <Xm/ScrollBar.h>
#include <Xm/DrawnB.h>
#include <Xm/Protocols.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Separator.h>
#include <Xm/DrawingA.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include "xmap.h"
#include "mode.h"
#include "debug.h"

#define DEBUG 1

extern jpeg_resync_to_restart;  //That is to force inclusion of libjpeg

XtAppContext app;

char *inifile;
Widget toplevel, main_w, toolbar, menubar, statusbar, draw_w;
Widget button_auto, button_comp, button_ucomp;
Widget status_1, status_2, status_3;
int insert_error;
GC status2GC;
GC status3GC;

void cal_dialog();
extern void _XEditResCheckMessages();

extern void MyMotionAction();
extern void MyEnterAction();
extern void MyLeaveAction();
extern void HlpMotionAction();
extern void HlpEnterAction();
extern void HlpLeaveAction();
extern void do_map_dialog();
extern void do_legend_dialog();
extern void do_pickdir_dialog();
extern void do_monitor_dialog();
extern void do_splash();
extern void do_about();
extern void track_start();
extern void track_stop();
extern void getlatlon();
extern void findplace();
extern void select_map_dialog();
extern void do_select_dialog();
extern void mw_move(Widget w, XMotionEvent * e);
extern void draw_up(Widget w, XMotionEvent * e);
extern void draw_down(Widget w, XMotionEvent * e);
extern void draw_motion(Widget w, XMotionEvent * e);
extern void draw_enter(Widget w, XMotionEvent * e);
extern void draw_leave(Widget w, XMotionEvent * e);
extern void draw_select(Widget w, XMotionEvent * e);
extern void open_log(), close_log(), track_from_log();
extern void print_map();
extern void popup_printer_tab();

#ifdef DND
extern void StampStartDrag();
#endif

extern void clear_marker_objects();
extern void clear_all_objects();
extern void do_relnotes();
extern void do_credits();
extern void save_objects();
extern void reload_objects();
extern void instrument_cb();
extern void speedometer_cb();
extern void aspeedometer_cb();
extern void position_cb();
extern void utmpos_cb();

extern void routes_cb();
extern void routen_cb();
extern void routev_cb();
extern void routec_cb();
extern void routee_cb();
extern void routex_cb();

static void new_file();
static void initial_new_file();

static XtActionsRec my_actions[] =
{
    {"MyMotionAction", (XtActionProc) MyMotionAction},
    {"MyEnterAction", (XtActionProc) MyEnterAction},
    {"MyLeaveAction", (XtActionProc) MyLeaveAction},
    {"HlpMotionAction", (XtActionProc) HlpMotionAction},
    {"HlpEnterAction", (XtActionProc) HlpEnterAction},
    {"HlpLeaveAction", (XtActionProc) HlpLeaveAction},
    {"draw_up", (XtActionProc) draw_up},
    {"draw_down", (XtActionProc) draw_down},
    {"draw_motion", (XtActionProc) draw_motion},
    {"draw_enter", (XtActionProc) draw_enter},
    {"draw_leave", (XtActionProc) draw_leave},
    {"draw_select", (XtActionProc) draw_select},
#ifdef DND
    {"StampStartDrag", (XtActionProc) StampStartDrag},
#endif
};

static char buttonTranslations[] =
"#override \
    <Motion> : MyMotionAction() \n\
    <Enter> : MyEnterAction() \n\
    <Leave> : MyLeaveAction() \n";

XtTranslations button_translations;

static char hlpbuttonTranslations[] =
"#override \
    <Motion> : HlpMotionAction() \n\
    <Enter> : HlpEnterAction() \n\
    <Leave> : HlpLeaveAction() \n";

XtTranslations hlpbutton_translations;

static char drawTranslations[] =
"   <Btn1Down>:   draw_down() \n\
    <Btn2Down>:   draw_select() \n\
    <Btn3Down>:   draw_select() \n\
    <BtnUp>:     draw_up() \n\
    <Motion>:     draw_motion() \n\
    <Enter>:      draw_enter() \n\
    <Leave>:      draw_leave()\n";

XtTranslations draw_translations;



/* This file only */
static char *mapdir;
static char *vmap_root;
static GC status1GC;

void exit_cb();
void toolbar_cb();
void toolhelp_cb();
void menubar_cb();
void grid_cb();
void statusbar_cb();
extern void draw_cb();
extern void redraw_cb();
extern void resize_cb();
extern void st3_redraw_cb();
extern char *get_tip(Widget w);

#ifdef DND
Atom PIXMAP_INDEX;
#endif

extern String cheapcolor;
extern String fallback_resources[];

/* options is defined in prefs.h */
OptionsRec options;

#define Offset(field) XtOffsetOf(OptionsRec, field)

XtResource resources[] =
{
    {"quiet", "Quiet", XtRBoolean, sizeof(Boolean),
     Offset(quiet), XtRImmediate, (XtPointer) FALSE},
    {"menubar", "Menubar", XtRBoolean, sizeof(Boolean),
     Offset(menubar), XtRImmediate, (XtPointer) TRUE},
    {"toolbar", "Toolbar", XtRBoolean, sizeof(Boolean),
     Offset(toolbar), XtRImmediate, (XtPointer) TRUE},
    {"toohelp", "Toolhelp", XtRBoolean, sizeof(Boolean),
     Offset(toolhelp), XtRImmediate, (XtPointer) FALSE},
    {"statusbar", "Statusbar", XtRBoolean, sizeof(Boolean),
     Offset(statusbar), XtRImmediate, (XtPointer) TRUE},
    {"grids", "Grids", XtRBoolean, sizeof(Boolean),
     Offset(grids), XtRImmediate, (XtPointer) FALSE},
    {"instruments", "Instruments", XtRBoolean, sizeof(Boolean),
     Offset(instruments), XtRImmediate, (XtPointer) FALSE},
    {"speedometer", "Speedometer", XtRBoolean, sizeof(Boolean),
     Offset(speedometer), XtRImmediate, (XtPointer) FALSE},
    {"aspeedometer", "Aspeedometer", XtRBoolean, sizeof(Boolean),
     Offset(aspeedometer), XtRImmediate, (XtPointer) FALSE},
    {"position", "Position", XtRBoolean, sizeof(Boolean),
     Offset(position), XtRImmediate, (XtPointer) FALSE},
    {"utmpos", "Utmpos", XtRBoolean, sizeof(Boolean),
     Offset(utmpos), XtRImmediate, (XtPointer) FALSE},
    {"nicecol", "Nicecol", XtRBoolean, sizeof(Boolean),
     Offset(nicecol), XtRImmediate, (XtPointer) TRUE},
    {"pdebug", "Pdebug", XtRBoolean, sizeof(Boolean),
     Offset(pdebug), XtRImmediate, (XtPointer) FALSE},
    {"useutc", "Useutc", XtRBoolean, sizeof(Boolean),
     Offset(useutc), XtRImmediate, (XtPointer) FALSE},
};

static XrmOptionDescRec cmd_options[] =
{
    {"-toolbar", "*toolbar", XrmoptionNoArg, "False"},
    {"+a", "*toolbar", XrmoptionNoArg, "True"},
    {"-a", "*toolbar", XrmoptionNoArg, "False"},
    {"+toolbar", "*toolbar", XrmoptionNoArg, "True"},
    {"-font", "*font", XrmoptionSepArg, "fixed"},
    {"-mapfile", "*MapFile", XrmoptionSepArg, NULL},
    {"-f", "*MapFile", XrmoptionSepArg, NULL},
    {"-in", "*MapFile", XrmoptionSepArg, NULL},
};

int unit = ENGLISH;
char *unit_names[] =
{"unknown", "english", "metric", "nautical"};
char *on_off[] =
{"off", "on"};

char *sform[] =
{"dddmm'ss.s\"", "dddmm.mmm'", "ddd.ddddd", "UTM"};

char *get_sform(int form) 
{
  return sform[form];
}

void ch_title(char *s)
{
    Arg args[1];
    char bufp[256];

    snprintf(bufp, 255, "Hugo Map Viewer - [%s]", s);
    XtSetArg(args[0], XmNtitle, bufp);
    XtSetValues(toplevel, args, 1);
}

MenuItem file_items[] =
{
    {"Open...", &xmPushButtonWidgetClass, 'O', "Ctrl<Key>O", "Ctrl+O", select_map_dialog, 0, NULL},
    {"Save", &xmPushButtonWidgetClass, 'S', "Ctrl<Key>S", "Ctrl+S", save_objects, 0, NULL},
    {"Import Markers", &xmPushButtonWidgetClass, 'I', NULL, NULL, import_markers, 0, NULL},
    {"Import Map", &xmPushButtonWidgetClass, 'M', NULL, NULL, import_map, 0, NULL},
    {"Discard Changes", &xmPushButtonWidgetClass, 'D', NULL, NULL, reload_objects, 0, NULL},
    {"sep1", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Print Map", &xmPushButtonWidgetClass, 'P', "Ctrl<Key>P", "Ctrl+P", print_map, 0, NULL},
    {"Printer Setup...", &xmPushButtonWidgetClass, 'e', NULL, NULL, popup_printer_tab, 0, NULL},
    {"sep2", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Exit", &xmPushButtonWidgetClass, 'x', NULL, NULL, exit_cb, 0, NULL},
    {NULL},
};

MenuItem edit_items[] =
{
    {"Clear Draw Objects", &xmPushButtonWidgetClass, 'C', NULL, "F3", clear_all_objects, 0, NULL},
    {"Clear Imported Objects", &xmPushButtonWidgetClass, 'I', NULL, NULL, clear_marker_objects, 0, NULL},
    {"sep1", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Calibrate Map", &xmPushButtonWidgetClass, 'M', NULL, NULL, cal_dialog, 0, NULL},
    {"sep2", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Preferences...", &xmPushButtonWidgetClass, 'P', NULL, NULL, do_select_dialog, 0, NULL},
    {NULL},
};

MenuItem search_items[] =
{
    {"Locate by Position", &xmPushButtonWidgetClass, 'L', NULL, NULL, getlatlon, 0, NULL},
    {"Locate Place Name", &xmPushButtonWidgetClass, 'P', NULL, NULL, findplace, 0, NULL},
    {NULL},
};

MenuItem view_items[] =
{
    {"Show Instruments", &xmToggleButtonWidgetClass, 'I', "Ctrl<Key>I", "Ctrl+I", instrument_cb, 0, NULL},
    {"sep1", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Digital Speedometer", &xmToggleButtonWidgetClass, 'D', NULL, NULL, speedometer_cb, 0, NULL},
    {"Analog Speedometer", &xmToggleButtonWidgetClass, 'S', NULL, NULL, aspeedometer_cb, 0, NULL},
    {"Position (Lat/Lon)", &xmToggleButtonWidgetClass, 'A', NULL, NULL, position_cb, 0, NULL},
    {"Position (UTM)", &xmToggleButtonWidgetClass, 'C', NULL, NULL, utmpos_cb, 0, NULL},
    {"sep2", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Director", &xmPushButtonWidgetClass, 'r', NULL, NULL, do_pickdir_dialog, 0, NULL},
    {"sep3", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"In 1 Mag", &xmPushButtonWidgetClass, 'I', NULL, "PgDn", zoomin, &mapdata, NULL},
    {"Out 1 Mag", &xmPushButtonWidgetClass, 'O', NULL, "PgUp", zoomout, &mapdata, NULL},
    {"Out 4 Mag", &xmPushButtonWidgetClass, '4', NULL, "Home", zoomout4, &mapdata, NULL},
    {NULL},
};

MenuItem route_items[] =
{
    {"Select Route...", &xmPushButtonWidgetClass, 'S', "Ctrl<Key>R", "Ctrl+R", routes_cb, 0, NULL},
    {"New Route...", &xmPushButtonWidgetClass, 'N', NULL, NULL, routen_cb, 0, NULL},
    {"Edit Current Route", &xmPushButtonWidgetClass, 'R', NULL, NULL, routee_cb, 0, NULL},
    {"Execute Current Route", &xmPushButtonWidgetClass, 'E', "Ctrl<Key>E", "Ctrl+E", routex_cb, 0, NULL},
    {"Save Current Route", &xmPushButtonWidgetClass, 'a', NULL, NULL, routev_cb, 0, NULL},
    {"sep1", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Clear Current Route", &xmPushButtonWidgetClass, 'C', NULL, NULL, routec_cb, 0, NULL},
    {NULL},
};

MenuItem option_items[] =
{
    {"Menubar", &xmToggleButtonWidgetClass, 'M', NULL, NULL, menubar_cb, 0, NULL},
    {"Toolbar", &xmToggleButtonWidgetClass, 'T', NULL, NULL, toolbar_cb, 0, NULL},
    {"Status Bar", &xmToggleButtonWidgetClass, 'S', NULL, NULL, statusbar_cb, 0, NULL},
    {"Tool Help", &xmToggleButtonWidgetClass, 'H', NULL, NULL, toolhelp_cb, 0, NULL},
    {"sep1", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Legend", &xmPushButtonWidgetClass, 'L', NULL, NULL, do_legend_dialog, 0, NULL},
    {"Overview Map", &xmPushButtonWidgetClass, 'O', NULL, NULL, 0, 0, NULL},
    {"sep2", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Map Features...", &xmPushButtonWidgetClass, 'M', NULL, NULL, do_map_dialog, NULL},
    {"Show Grids", &xmToggleButtonWidgetClass, 'G', NULL, NULL, grid_cb, 0, NULL},
    {NULL},
};

MenuItem gps_items[] =
{
    {"Start Tracking", &xmPushButtonWidgetClass, 'S', NULL, "F12", track_start, 0, NULL},
    {"Stop Tracking", &xmPushButtonWidgetClass, 'p', NULL, NULL, track_stop, 0, NULL},
    {"sep1", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Monitor GPS...", &xmPushButtonWidgetClass, 'G', NULL, NULL, do_monitor_dialog, 0, NULL},
    {"sep2", &xmSeparatorWidgetClass, 0, NULL, NULL, 0, 0, NULL},
    {"Track to Log...", &xmPushButtonWidgetClass, 'T', NULL, NULL, open_log, 0, NULL},
    {"Stop Tracking to Log", &xmPushButtonWidgetClass, 'o', NULL, NULL, close_log, 0, NULL},
    {"Track from Log...", &xmPushButtonWidgetClass, 'r', NULL, NULL, track_from_log, 0, NULL},
    {NULL},
};


MenuItem help_items[] =
{
    {"Index", &xmPushButtonWidgetClass, 'I', NULL, NULL, do_help, 0, NULL},
    {"Release Notes", &xmPushButtonWidgetClass, 'R', NULL, NULL, do_relnotes, 0, NULL},
    {"Credits", &xmPushButtonWidgetClass, 'C', NULL, NULL, do_credits, 0, NULL},
    {"About", &xmPushButtonWidgetClass, 'A', NULL, NULL, do_about, 0, NULL},
    {NULL},
};




MenuItem bar_items[] =
{
  {"file", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, file_items},
  {"edit", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, edit_items},
    {"search", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, search_items},
  {"view", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, view_items},
  {"route", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, route_items},
    {"options", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, option_items},
    {"gps", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, gps_items},
  {"help", &xmPushButtonWidgetClass, 0, NULL, NULL, 0, 0, help_items},
    {NULL},
};


Widget make_menubar(Widget parent)
{
    Widget w1, w2, w3, w4, w5;

    menubar = BuildMenu(parent, XmMENU_BAR, "menuBar", 0, bar_items);

    XtManageChild(menubar);

    /* make the Help button known */
    if ( (w1=bar_items[7].widget) )
	XtVaSetValues(menubar, XmNmenuHelpWidget, w1, NULL);
    else fprintf(stderr,"Wrong order 0\n");

    if ((w1 = option_items[0].widget)
	&& (w2 = option_items[1].widget)
	&& (w3 = option_items[2].widget)
	&& (w4 = option_items[3].widget)
	&& (w5 = option_items[9].widget)) {

	XmToggleButtonSetState(w1, options.menubar, FALSE);
	XmToggleButtonSetState(w2, options.toolbar, FALSE);
	XmToggleButtonSetState(w3, options.statusbar, FALSE);
	XmToggleButtonSetState(w4, options.toolhelp, FALSE);
	XmToggleButtonSetState(w5, options.grids, FALSE);
    }
    else fprintf(stderr,"Wrong order 1\n");

    // Handle Instruments
    if ((w1 = view_items[0].widget)
	&& (w2 = view_items[2].widget)
	&& (w3 = view_items[3].widget)
	&& (w4 = view_items[4].widget)
	&& (w5 = view_items[5].widget)) {

	XmToggleButtonSetState(w1, options.instruments, FALSE);
	XmToggleButtonSetState(w2, options.speedometer, FALSE);
	XmToggleButtonSetState(w3, options.aspeedometer, FALSE);
	XmToggleButtonSetState(w4, options.position, FALSE);
	XmToggleButtonSetState(w5, options.utmpos, FALSE);
    }
    else fprintf(stderr,"Wrong order 2\n");

    return menubar;
}


void handle_toolbar_visibility()
{
    if (options.toolbar == FALSE) {
	XtUnmanageChild(toolbar);
	XtVaSetValues(draw_w,
		      XmNtopAttachment, XmATTACH_FORM,
		      NULL);
    } else {
	XtManageChild(toolbar);
	XtVaSetValues(draw_w,
		      XmNtopAttachment, XmATTACH_WIDGET,
		      XmNtopWidget, toolbar,
		      NULL);
    }
}

void handle_menubar_visibility()
{
    if (options.menubar == FALSE) {
	XtUnmanageChild(menubar);
//      XtVaSetValues(draw_w,
	//                    XmNtopAttachment, XmATTACH_FORM,
	//                    NULL);
    } else {
	XtManageChild(menubar);
//      XtVaSetValues(draw_w,
	//                    XmNtopAttachment, XmATTACH_WIDGET,
	//                    XmNtopWidget, toolbar,
	//                    NULL);
    }
}

void handle_statusbar_visibility()
{
    if (options.statusbar == FALSE) {
	XtUnmanageChild(statusbar);
	XtVaSetValues(draw_w,
		      XmNbottomAttachment, XmATTACH_FORM,
		      NULL);
    } else {
	XtManageChild(statusbar);
	XtVaSetValues(draw_w,
		      XmNbottomAttachment, XmATTACH_WIDGET,
		      XmNbottomWidget, statusbar,
		      NULL);
    }
}

void status_field(Widget * w, char *name, int left, int right)
{
    *w = XtVaCreateManagedWidget(name, xmTextFieldWidgetClass, statusbar,
				 XmNcursorPositionVisible, False,
				 XmNeditable, False,
				 XmNmarginHeight, 1,
				 XmNhighlightThickness, 0,
				 XmNshadowThickness, 1,
				 XmNleftAttachment, XmATTACH_POSITION,
				 XmNleftPosition, left,
				 XmNrightAttachment, XmATTACH_POSITION,
				 XmNrightPosition, right,
				 NULL);
}

void status_fieldf(Widget * w, char *name, int left, int right)
{
    Widget w0;

    w0 = XtVaCreateManagedWidget(name, xmFrameWidgetClass, statusbar,
				 XmNmarginHeight, 1,
				 XmNhighlightThickness, 0,
				 XmNshadowThickness, 1,
				 XmNleftAttachment, XmATTACH_POSITION,
				 XmNleftPosition, left,
				 XmNrightAttachment, XmATTACH_POSITION,
				 XmNrightPosition, right,
				 XmNtopAttachment, XmATTACH_FORM,
				 XmNbottomAttachment, XmATTACH_FORM,
				 NULL);

    *w = XtVaCreateManagedWidget(name, xmDrawingAreaWidgetClass, w0,
				 XmNmarginHeight, 1,
				 XmNhighlightThickness, 0,
				 XmNshadowThickness, 1,
				 XmNleftAttachment, XmATTACH_POSITION,
				 XmNleftPosition, left,
				 XmNrightAttachment, XmATTACH_POSITION,
				 XmNrightPosition, right,
				 XmNtopAttachment, XmATTACH_FORM,
				 XmNbottomAttachment, XmATTACH_FORM,
				 NULL);
}

Pixel get_pixel(Widget w, char *resource_value)
{
    Colormap colormap;
    Boolean status;
    XColor exact, color;

    colormap = DefaultColormapOfScreen(DefaultScreenOfDisplay(XtDisplay(w)));

    status = XLookupColor(XtDisplay(w), colormap, resource_value, &color, &exact);


    if (status == 0) {
	fprintf(stderr, "Unknown color: %s", resource_value);
	color.pixel = BlackPixelOfScreen(DefaultScreenOfDisplay(XtDisplay(w)));
    } else {
	myXAllocColor(XtDisplay(w), colormap, &color);
    }

    return (color.pixel);
}

int get_option(char *section, char *entry, char *def)
{
    char bufp[10];

    GetPrivateProfileString(section, entry, def, bufp, 5, inifile);
    if (strcasecmp(bufp, "off") == 0)
	return 0;
    if (strcasecmp(bufp, "false") == 0)
	return 0;
    if (strcasecmp(bufp, "not") == 0)
	return 0;
    if (strcasecmp(bufp, "no") == 0)
	return 0;
    return 1;
}

void get_units()
{
    char bufp[24];

    GetPrivateProfileString("General", "units", unit_names[ENGLISH],
			    bufp, 20, inifile);

    if (strcasecmp(unit_names[METRIC], bufp) == 0)
	unit = METRIC;
    else if (strcasecmp(unit_names[NAUTICAL], bufp) == 0)
	unit = NAUTICAL;
    else
	unit = ENGLISH;
}

void schedule_initial_read(char *file)
{
    XtAppAddTimeOut(app, 100L, (XtTimerCallbackProc) initial_new_file, file);
}

void schedule_map_update(char *file)
{
    XtAppAddTimeOut(app, 100L, (XtTimerCallbackProc) new_file, file);
}

static int check_inifile()
{
    struct stat sb;
    char bufp[256];

    // The file 'inifile' might not exist. Things just won't
    // work well without it. The whole vector map stuff will
    // break horribly. Try to get the template from COMMON_PATH.

    if (stat(inifile, &sb)) {
        snprintf(bufp, 255, "%s/hugo.ini", COMMON_PATH);
        if (stat(bufp, &sb)==0) {
            return copy_file(bufp, inifile);
        }       
        else {
            fprintf(stderr, "Warning: You need an initialization file\n"
                            "         named 'hugo.ini' in your home\n"
                            "         directory or in %s\n"
                            "         a temporary one will be created,\n"
                            "         vector maps will not work with it.\n",
                            COMMON_PATH);
        }
    }
    return 0;
}

main(argc, argv)
int argc;
char *argv[];
{
    Widget rc;
    XGCValues gcv;
    char *home = getenv("HOME");
    Atom delw;
    Display *dpy;
    char bufp[256];
    char *filename;

    if (!home)
	home = "./";

    snprintf(bufp, 255, "%s/hugo.ini", home);
    inifile = strdup(bufp);
    check_inifile();

    if (get_option("General", "nicecol", "on"))
	fallback_resources[0] = cheapcolor;

    /* initialize toolkit and create toplevel shell */
    toplevel = XtVaAppInitialize(&app, "Hugo",
			cmd_options, XtNumber(cmd_options), &argc, argv,
			fallback_resources, NULL);

    dpy = XtDisplay(toplevel);
    delw = XmInternAtom(dpy, "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(toplevel, delw,
                            (XtCallbackProc) exit_cb, (XtPointer) NULL);



    options.format = GetPrivateProfileInt("General", "ddd", 0, inifile);

    GetPrivateProfileString("Layout", "main", "600x400", bufp, 255, inifile);
    XtVaSetValues(toplevel, XmNgeometry, bufp, NULL);

    XtAddEventHandler(toplevel, (EventMask) 0, True,
		      _XEditResCheckMessages, NULL);

    XtAppAddActions(app, my_actions, XtNumber(my_actions));

    XtVaGetApplicationResources(toplevel, (XtPointer) & options,
			resources, XtNumber(resources), 
			NULL);

#ifdef DND
    PIXMAP_INDEX = XmInternAtom(dpy, "PIXMAP_INDEX", False);
#endif


    /* MainWindow for the application -- contains menubar
     * and ScrolledText/Prompt/TextField as WorkWindow.
     */
    main_w = XtVaCreateManagedWidget("main_w",
				xmMainWindowWidgetClass, toplevel,
				NULL);

    button_translations = XtParseTranslationTable(buttonTranslations);
    hlpbutton_translations = XtParseTranslationTable(hlpbuttonTranslations);

    options.grids = get_option("General", "grid", "off");
    options.toolbar = get_option("General", "toolbar", "on");
    options.statusbar = get_option("General", "statusbar", "on");
    options.toolhelp = get_option("General", "toolhelp", "off");
    options.nicecol = get_option("General", "nicecol", "on");
    options.useutc = get_option("General", "useutc", "on");

    // Handle the instruments
    options.instruments = get_option("General", "instruments", "off");
    options.speedometer = get_option("General", "speedometer", "on");
    options.aspeedometer = get_option("General", "analog_speedometer", "on");
    options.position = get_option("General", "digiposition", "on");
    options.utmpos = get_option("General", "utmposition", "off");
    get_units();

    menubar = make_menubar(main_w);

    rc = XtVaCreateWidget("work_area", xmFormWidgetClass, main_w, NULL);



    toolbar = XtVaCreateManagedWidget("toolbar", xmBulletinBoardWidgetClass, rc,
				      XmNresizePolicy, XmRESIZE_NONE,
				      XmNmarginHeight, 4,
				      XmNmarginWidth, 4,
				      XmNtopAttachment, XmATTACH_FORM,
				      XmNleftAttachment, XmATTACH_FORM,
				      XmNrightAttachment, XmATTACH_FORM,
				      XmNheight, 34,
				      XmNminHeight, 34,
				      XmNmaxHeight, 34,
				      NULL);


    draw_w = XtVaCreateWidget("draw_w",
			      xmDrawingAreaWidgetClass, rc,
			      XmNleftAttachment, XmATTACH_FORM,
			      XmNrightAttachment, XmATTACH_FORM,
			      XmNtranslations, XtParseTranslationTable(drawTranslations),
			      NULL);
#if 0
    draw_translations = XtParseTranslationTable(drawTranslations);
    XtOverrideTranslations(draw_w, draw_translations);
#endif

    register_canvas(draw_w);

    XtAddCallback(draw_w, XmNexposeCallback, redraw_cb, NULL);
    XtAddCallback(draw_w, XmNresizeCallback, resize_cb, NULL);

    statusbar = XtVaCreateWidget("statusbar", xmFormWidgetClass, rc,
				 XmNbottomAttachment, XmATTACH_FORM,
				 XmNleftAttachment, XmATTACH_FORM,
				 XmNrightAttachment, XmATTACH_FORM,
				 XmNfractionBase, 200,
				 NULL);

    status_field(&status_1, "text_1", 0, 120);
    status_field(&status_2, "text_2", 121, 150);
    status_fieldf(&status_3, "text_3", 151, 200);
    XtAddCallback(status_3, XmNexposeCallback, st3_redraw_cb, NULL);

    gcv.foreground = BlackPixelOfScreen(XtScreen(draw_w));
    status1GC = XCreateGC(XtDisplay(status_1),
	       RootWindowOfScreen(XtScreen(status_1)), GCForeground, &gcv);

    status2GC = XCreateGC(XtDisplay(status_2),
	       RootWindowOfScreen(XtScreen(status_2)), GCForeground, &gcv);

    status3GC = XCreateGC(XtDisplay(status_3),
	       RootWindowOfScreen(XtScreen(status_3)), GCForeground, &gcv);

    handle_toolbar_visibility();
    handle_statusbar_visibility();
    XtManageChild(draw_w);
    XtManageChild(rc);

    XmMainWindowSetAreas(main_w, menubar, NULL, NULL, NULL, rc);
    ch_title("none");

    XtRealizeWidget(toplevel);

    instrument_init();

    ch_mode(NAVIGATE,START);

    create_help_dialog();

    loadIcons();
    handle_instrument_visibility();

    snprintf(bufp, 255, "%s/maps", COMMON_PATH);
    GetPrivateProfileString("General", "mapdir", bufp, bufp, 255, inifile);
    mapdir = strdup(bufp);

    snprintf(bufp, 255, "%s/maps/vmap/", COMMON_PATH);
    GetPrivateProfileString("General", "vmapdir", bufp, bufp, 255, inifile);
    vmap_root = strdup(bufp);

    verify_mapdirs();
    
    GetPrivateProfileString("Maps", "current", "world.ini", bufp, 255, 
			    inifile);
    filename = strdup(bufp);

    do_splash(); // Draw the splash screen
    set_cursor(1);
    install_sig_handlers();
    start_heartbeat();
    schedule_initial_read(filename);

    XtAppMainLoop(app);

    return -1;
}

char *get_mapdir()
{
  return mapdir;
}

char *get_vmap_root()
{
  return vmap_root;
}

/*
 * We want to save as much state as we can.
 */
void exit_cb()
{
    Position x, y;
    Dimension width, height;
    char bufp[256];

    XtVaGetValues(toplevel, XmNx, &x, XmNy, &y,
		  XmNwidth, &width, XmNheight, &height,
		  NULL);
    snprintf(bufp, 255, "%dx%d+%d+%d", width, height, x, y);
    WritePrivateProfileString("Layout", "main", bufp, inifile);

    // Handle the instruments
    WritePrivateProfileString("General", "instruments", on_off[(int)options.instruments], inifile);
    WritePrivateProfileString("General", "speedometer", on_off[(int)options.speedometer], inifile);
    WritePrivateProfileString("General", "analog_speedometer", on_off[(int)options.aspeedometer], inifile);
    WritePrivateProfileString("General", "digiposition", on_off[(int)options.position], inifile);
    WritePrivateProfileString("General", "utmposition", on_off[(int)options.utmpos], inifile);

    save_legend_properties();
    save_pickdir_properties();
    save_help_properties();
    save_instrument_properties();
    close_log();
    save_objects();
    WriteOutProfiles();

    exit(0);
}

static void initial_new_file(char *file)
{
    eval_name(file);
    new_map(file);
    /* We always strdup() to get this */
    free(file);
    read_objects();
    redraw_objects();
    redraw_location();
}

static void new_file(char *file)
{
    new_map(file);
    /* We always strdup() to get this */
    free(file);
    redraw_objects();
    redraw_location();
}

void toolbar_cb(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct * state)
{
    options.toolbar = state->set;
    handle_toolbar_visibility();
    WritePrivateProfileString("General", "toolbar", on_off[state->set], inifile);
    WriteOutProfiles();
}

void toolhelp_cb(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct * state)
{
    options.toolhelp = state->set;
    WritePrivateProfileString("General", "toolhelp", on_off[state->set], inifile);
    WriteOutProfiles();
}

void menubar_cb(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct * state)
{
//    options.menubar = state->set;
    handle_toolbar_visibility();
//    WritePrivateProfileString("General", "menubar", on_off[state->set], inifile);
    //    WriteOutProfiles();
}

void statusbar_cb(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct * state)
{
    options.statusbar = state->set;
    handle_statusbar_visibility();
    WritePrivateProfileString("General", "statusbar", on_off[state->set], inifile);
    WriteOutProfiles();
}

void grid_cb(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct * state)
{
    options.grids = state->set;
//    handle_grid_visibility();
    WritePrivateProfileString("General", "grid", on_off[state->set], inifile);
    WriteOutProfiles();
}

int qconf_graticule()
{
    return options.grids;
}

void beep(void)
{
    XBell(XtDisplay(toplevel), 0);
}
